#include <netinet/in.h>
#include <linux/netfilter.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <libnetfilter_conntrack/libnetfilter_conntrack.h>
#include <libnetfilter_conntrack/libnetfilter_conntrack_tcp.h>
#include <netinet/ip.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <netinet/ip.h>
#include <fcntl.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#define __FAVOR_BSD
#include <netinet/tcp.h>

#undef _POSIX_SOURCE    

#include "tcpcrypt_divert.h"
#include "tcpcryptd.h"

static struct nfq_handle    *_h;
static struct nfq_q_handle  *_q;
static unsigned int	    _mark;
static struct nfct_handle   *_ct;

static struct ct_arg {
	struct sockaddr_in	*ct_to;
	struct ip		*ct_ip;
	int			*ct_flags;
	int			ct_rc;
} _ct_arg;

static int packet_input(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg,
              		struct nfq_data *nfa, void *data)
{
	divert_cb cb = (divert_cb) data;
	unsigned char *d;
	int len;
	int rc;
	unsigned int id;
	struct nfqnl_msg_packet_hdr *ph = nfq_get_msg_packet_hdr(nfa);
	struct ip *ip;
	int flags = 0;
	struct timeval tv;
	int rlen = 0;
	void *rdata = NULL;

	len = nfq_get_payload(nfa, &d);
	if (len < 0)
		err(1, "nfq_get_payload()");

	if (nfq_get_indev(nfa))
		flags |= DF_IN;

	if (nfq_get_timestamp(nfa, &tv) == 0)
		set_time(&tv);
	else {
		static int warn = 0;

		if (!warn && !_conf.cf_disable_timers)
			xprintf(XP_ALWAYS, "No timestamp provided in packet"
			                   " - expect low performance due to"
					   " calls to gettimeofday\n");
		warn = 1;	
	}

	rc = cb(d, len, flags);

	id = ntohl(ph->packet_id);

	switch (rc) {
	case DIVERT_MODIFY:
		ip    = (struct ip*) d;
		rlen  = ntohs(ip->ip_len);
		rdata = d;
		/* fallthrough */
	case DIVERT_ACCEPT:
		if (_mark) {
			unsigned int mark = 0;

			assert((mark & _mark) == 0);
			nfq_set_verdict_mark(qh, id, NF_REPEAT,
					     htonl(_mark | mark),
					     rlen, rdata);
		} else
			nfq_set_verdict(qh, id, NF_ACCEPT, rlen, rdata);
		break;

	case DIVERT_DROP:
		nfq_set_verdict(qh, id, NF_DROP, 0, NULL);
		break;

	default:
		printf("Unknown verdict %d\n", rc);
		abort();
	}

	return 0;
}

static int conntrack_find(enum nf_conntrack_msg_type type,
			  struct nf_conntrack *ct,
			  void *data)
{
	struct ct_arg *arg = data;
	int *flags = arg->ct_flags;
	struct sockaddr_in *to = arg->ct_to;
	struct ip *ip = arg->ct_ip;
	struct tcphdr *tcp = (struct tcphdr*) ((unsigned long) ip
                                               + ip->ip_hl * 4);

	if (arg->ct_rc == 0)
		return NFCT_CB_CONTINUE;

#if 0
	char buf[1024];
	nfct_snprintf(buf, sizeof(buf), ct, type, NFCT_O_DEFAULT, 0);
	printf("YO [%s]\n", buf);
	return NFCT_CB_CONTINUE;
#endif

	if (nfct_get_attr_u32(ct, ATTR_IPV4_SRC) != ip->ip_src.s_addr)
		return NFCT_CB_CONTINUE;

	if (nfct_get_attr_u16(ct, ATTR_PORT_SRC) != tcp->th_sport)
		return NFCT_CB_CONTINUE;

	switch (nfct_get_attr_u8(ct, ATTR_TCP_STATE)) {
	case TCP_CONNTRACK_SYN_RECV:
		*flags = DF_IN;
		break;

	case TCP_CONNTRACK_SYN_SENT:
		*flags = 0;
		break;

	default:
		return NFCT_CB_CONTINUE;
	}

	to->sin_addr.s_addr = nfct_get_attr_u32(ct, ATTR_ORIG_IPV4_DST);
	to->sin_port        = nfct_get_attr_u16(ct, ATTR_ORIG_PORT_DST);

	arg->ct_rc = 0;

	return NFCT_CB_CONTINUE;
}

static void conntrack_open(void)
{
	if (!(_ct = nfct_open(CONNTRACK, 0)))
		err(1, "nfct_open()");

	nfct_callback_register(_ct, NFCT_T_ALL, conntrack_find, &_ct_arg);
}

static int linux_open(int port, divert_cb cb)
{
	unsigned int bufsize = 1024 * 1024 * 1;
	unsigned int rc;
	char *m;
	int fd, flags;

        _h = nfq_open();
        if (!_h)
                err(1, "nfq_open()");

	rc = nfnl_rcvbufsiz(nfq_nfnlh(_h), bufsize);
	if (rc != bufsize)
		xprintf(XP_DEBUG, "Buffer size %u wanted %u\n", rc, bufsize);

	/* reset in case of previous crash */
	if (nfq_unbind_pf(_h, AF_INET) < 0)
		err(1, "nfq_unbind_pf()");

        if (nfq_bind_pf(_h, AF_INET) < 0)
                err(1, "nfq_bind_pf()");

        _q = nfq_create_queue(_h, port, packet_input, cb);
        if (!_q)
                err(1, "nfq_create_queue()");

        if (nfq_set_mode(_q, NFQNL_COPY_PACKET, 0xffff) < 0)
                err(1, "nfq_set_mode()");

	if (nfq_set_queue_maxlen(_q, 10000) < 0)
		err(1, "nfq_set_queue_maxlen()");

       	xprintf(XP_DEFAULT,
		"Divert packets using iptables -j NFQUEUE --queue-num %d\n",
                port);

	m = driver_param(0);
	if (m) {
		_mark = strtoul(m, NULL, 16);
		xprintf(XP_DEFAULT, "Also, add -m mark --mark 0x0/0x%x\n",
			_mark);
	}

	fd = nfq_fd(_h);

	flags = fcntl(fd, F_GETFL);
	if (flags == -1)
		err(1, "fcntl()");

	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
		err(1, "fcntl()");

	raw_open();
	conntrack_open();

	return fd;
}

static void linux_close(void)
{
        if (_q)
                nfq_destroy_queue(_q);

        if (_h)
                nfq_close(_h);

	if (_ct)
		nfct_close(_ct);
}

static void linux_next_packet(int s)
{
	char buf[2048];
	int rc;

	rc = read(s, buf, sizeof(buf));
	if (rc == -1) {
		if (errno == ENOBUFS) {
			printf("FUCK - we're dropping packets\n");
			return;
		}

		err(1, "read(divert) %d", errno);
	}

	if (rc == 0)
		errx(1, "EOF");

	nfq_handle_packet(_h, buf, rc);
}

static int linux_orig_dest(struct sockaddr_in *to, struct ip *ip, int *flags)
{
	int family = AF_INET;
	struct ct_arg *arg = &_ct_arg;

	memset(arg, 0, sizeof(*arg));

	arg->ct_to    = to;
	arg->ct_ip    = ip;
	arg->ct_flags = flags;
	arg->ct_rc    = -1;

	/* XXX have specific filter */
	if (nfct_query(_ct, NFCT_Q_DUMP, &family) < 0)
		err(1, "nfct_query()");

	return arg->ct_rc;
}

struct divert *divert_get(void)
{
	static struct divert _divert_linux = {
		.open		= linux_open,
		.next_packet	= linux_next_packet,
		.close		= linux_close,
		.inject		= raw_inject,
		.orig_dest	= linux_orig_dest,
	};

        if (_conf.cf_rdr) {
                struct divert *pcap = divert_get_pcap();
                _divert_linux.inject = pcap->inject;
        }

	return &_divert_linux;
}
